#pragma once

#include <functional>
#include <memory>
#include <atomic>
#include <utility>

namespace MultiThreading {
    class TaskBase;

    template <typename T>
    class Task;
}

class MultiThreading::TaskBase {
protected:
    std::atomic<bool> complete = false;
public:
    /// @returns whether the task is finished.
    [[nodiscard]] bool Complete() const { return complete; }
    virtual void Run() = 0;
public:
    TaskBase() = default;
};

template <typename T>
class MultiThreading::Task : public TaskBase {
private:
    T* result = nullptr;
    std::function<T()> callable = nullptr;
private:
    explicit Task(std::function<T()> callable, T* result = nullptr) : TaskBase(), result(result), callable(callable) {}
public:
    void Run() final { result ? *result = callable() : callable(); complete = true; }
public:
    /// Create a Task. This is non-negotiable, This ensures that there's no issues related to the stack frame the task is on being popped before or while the job runs.
    /// @param callable The function to run, *usually a lambda or std::bind*
    /// @param result If your function returns a value, This is where you want it to go. nullptr for no return value or you don't want it.
    /// @note this is shared_ptr so you don't have to delete it.
    static std::shared_ptr<Task<T>> Create(std::function<T()> callable, T* result = nullptr) { return std::shared_ptr<Task<T>>(new Task<T>(callable, result)); };
    ~Task() = default;
};

// Special case for if the task is void return type because templates are weird.
template <>
class MultiThreading::Task<void> : public TaskBase {
private:
    std::function<void()> callable = nullptr;
private:
    explicit Task(std::function<void()> callable) : TaskBase(), callable(std::move(callable)) {}
public:
    void Run() final { callable(); complete = true; }

    /// Create a Task. This is non-negotiable, This ensures that there's no issues related to the stack frame the task is on being popped before or while the job runs.
    /// @param callable The function to run, *usually a lambda or std::bind*
    /// @note this is shared_ptr so you don't have to delete it.
    static std::shared_ptr<Task<void>> Create(const std::function<void()>& callable) { return std::shared_ptr<Task<void>>(new Task<void>(callable)); }
};